# Copyright (C) 2022, Hewlett Packard Enterprise, Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
"""Utility module for Hewlett Packard Enterprise Driver."""

VERSION = '1.0.0'
CI_WIKI_NAME = 'HPE_XP_Storage_CI'
PARAM_PREFIX = 'hpexp'
VENDOR_NAME = 'Hewlett Packard Enterprise'
DRIVER_PREFIX = 'HPEXP'
DRIVER_FILE_PREFIX = 'hpe_xp'
TARGET_PREFIX = 'HPEXP-'
HDP_VOL_ATTR = 'THP'
HDT_VOL_ATTR = 'ST'
NVOL_LDEV_TYPE = 'THP V-VOL'
TARGET_IQN_SUFFIX = '.hpexp-target'
PAIR_ATTR = 'FS'
