{
    "backup": {
        "container": null,
        "description": "Test backup",
        "name": "backup001",
        "volume_id": "%(volume_id)s",
        "incremental": false,
        "snapshot_id": null,
        "force": false
    }
}
